/* SPDX-License-Identifier: GPL-2.0-only */
#ifndef __LABWC_SPAWN_H
#define __LABWC_SPAWN_H

/**
 * spawn_async_no_shell - execute asyncronously
 * @command: command to be executed
 */
void spawn_async_no_shell(char const *command);

#endif /* __LABWC_SPAWN_H */
